:- module( lib, [
                    lib/1, lib/2   % +Repo[, +Opts]
        ] ).

                     % lib_suggests/1+2,  % fixme: feature()
                     % lib_expects/1+2,
                     % lib_init/1

:- ensure_loaded( library(prolog_pack) ).
    % query_pack_server/3

:- ensure_loaded( '../src/lib_auxil' ).
:- ensure_loaded( '../src/lib_attach').
:- ensure_loaded( '../src/lib_load'  ).
:- ensure_loaded( '../src/lib_type'  ).
:- ensure_loaded( '../src/lib_homonyms'  ).
:- ensure_loaded( '../src/lib_suggests'  ).
:- ensure_loaded( '../src/lib_expects'  ).
:- ensure_loaded( '../src/lib_init'  ).
:- ensure_loaded( '../src/lib_message'  ).


:- dynamic( lib_tables:lib_repo/4 ).            % +Repo, +Type, +Root, +Load 
:- dynamic( lib_tables:lib_repo_index/2 ).      % +Repo, +IdxFile
:- dynamic( lib_tables:lib_repo_homonyms/2 ).   % +Repo, +SrcDir
:- dynamic( lib_tables:lib_context/2 ).         % +Ctx, +Root
:- dynamic( lib_tables:lib_index/4 ).           % +Pa, +Pn, +Repo, +File
                                                % records all loaded indices
:- dynamic( lib_tables:lib_homonym/3 ).         % +Stem, +Repo, +File
                                                % records all loaded homonyms
:- dynamic( lib_tables:lib_loaded_index/2 ).    % +Repo, +File
                                                % keeps track of loaded index files
:- dynamic( lib_tables:lib_loaded_homonyms/2 ). % 
:- dynamic( lib_tables:lib_attached_indices/2 ).% +Ctx, Repo
:- dynamic( lib_tables:lib_attached_homonyms/2 ).% +Ctx, Repo


% fixme: user defined ones
lib_src_sub_dir( src ).
lib_src_sub_dir( 'src/lib' ).
lib_src_sub_dir( 'src/auxil' ).

/** <module> Predicate based code development.

 This pack implements methods for loading code into SWI Prolog programs.
 One of the major innovations the library introduces is that of allowing
 partial loading of packs. That is, if only a specific predicate is required
 from a =|pack(lib)-aware|= pack, only that and its dependent code will be loaded.

 Although this library, _pack(lib)_, contains a number of involved features
 it can also be used as a straight forward shorthand, 
 replacement for use_module(library(Lib)).

 ==
 ?- lib(Atomic).
 ==
 is equivelant to =|use_module(library(Atomic))|= if Atomic is a system library
 or an installed Pack, while it will interogate the SWI pack server 
 for matching packs if Atomic is atomic and not an installed pack.

In addition the library allows for loading with initializations turned off.

---++ Repositories

Code is managed in _repositories_ (also _repo_) that can be either packs or libs. 

A _pack_ is a unit of programs as managed by built-in SWI package manager 
(=|library(prolog_pack)|=). A _lib_ (_library_) is a directory containing a
number of program files.

pack(lib) supports a number of ways to organise your code and load it, but
it comes to its own when code s organised as predicate-in-a-file fashion.
In this mode of development a predicate such as kv_decompose/3
would be defined on file kv_decompose.pl which will only containing code for defining
this predicate, with the possible exception of helper predicates that are
too specific to be of outside interest.
==
kv_decompose( [], [], [] ).
kv_decompose( [K-V|T], [K|Tk], [V|Tv] ) :-
	kv_decompose( T, Tk, Tv ).
==

Lib code is considered as coming from the special pack _user_.

---++ Code-tables

Associated with each repository are 2 types of code-tables: (code-)_indices_ and (file-)_locators_.

A code-index maps a predicate identifier along with its source repo to
an absolute file name of the source that defines it.
Indices are of the form:

==
lib_tables:lib_index( Pname, Parity, Repo, AbsFile ).
==

File locators store all filenames in a repository. 
These can be named matched predicate names that need to be loaded.
pack(lib) can be directed to assume that files from a specific 
repository exhibit this _homonyms_ property.
Locators are of the form:
==
lib_tables:lib_homonym( Stem, Repo, AbsFile ).
==

For each index file loaded for a repository, the following is asserted:
==
lib_tables:lib_loaded_indices(Repo,File)
==

and for each locator 
==
lib_tables:lib_loaded_homonyms(Repo,Stem,File)
==

When loading a repository the user can choose whether to load
indices and locators independently.

---++ Loading source code

During the process of loading code into memory, lib/1 and /2 directives are used
to locate code to which the specific code depends. 

There are three main categories of operations: 
  * load code from explicitly identified repository
  * attach a repository for implicit loading
  * load code from attached repositories

These operations are all specific to the loading context. This is achieved by 
creating meta-predicates that identify which part of the repository base each
context has access to.

Attachment of repository is registered via =|lib_tables:lib_attached_indices(To,PackIG)|=
and =|lib_tables:lib_attached_homonyms(To,PackFG)|=.
==
lib_tables:lib_attached_indices(bims,options).
lib_tables:lib_attached_homonyms(bims,false).
==
attaches the indices but not the file locators.

Since all code from directory-libs load to a single module (_user_), loading 
code has either access to all such code, or to none.

---++ Conventions

Packs are expected to have matching top directories and main files. The main file
of a pack should be within top directory _prolog/_. (The directory convention 
is set by library(prolog_pack)). For example for pack _bims_ the following file should exist 
in packs directory:

==
bims/prolog/bims.pl
==

For packs the main code directory is _src/_. Additionally _src/lib_ and _src/auxil_
are treated as code directories.

---++ Internals

Variables
  * Repo
    repository
  * Pack
    prolog pack (installed or locally addressed)
  * Lib
    directory containing code 
  * Root 
    absolute reference to the root directory of a Repo
  * Pn 
    predicate name
  * Idx 
    library index term
  * Hmn
    library file homonym term
  * Pa 
    predicate arity
  * Cxt 
    context module

Predicate names
  * index
  * file

---++ Pack info

This is a complete re-write of pack(requires) v1.1.

Listens to =|debug(lib)|=.

@author nicos angelopoulos
@version  1.0 2017/3/6
@see http://stoics.org.uk/~nicos/sware/lib

*/

lib_defaults( pack, [load(true),index(true),homonym(true),type(pack),mode(self)] ).
lib_defaults( lib, [load(false),index(true),homonym(true),type(lib),mode(self)] ).

/**  lib( +Operand ).
     lib( +Operand, +Opts )

Loads code or/and indices of Repo into the current context.

When Repo =|homonym(Repository)|= then only the homonims of local dir
(adjusted for pack dir structure) are added to as coming from Repository.

Operands
  * SysLibrary
    An installed library (atomic). Is loaded with use_module( library(SysLibrary) ).
  * Pack
    A pack known to SWI. If pack is not installed then the server is contacted to look
    for name-matching packs that can be installed. If there is at least
    one matching pack, it can be installed interactively.
  * lib 
    Declares a library directory. The default options for libs are
      * load(Load=false)
      * index(Idx=true)
      * homonyms(Hmns=true)
      * type(Type=lib)

  * alias(Dir)
    As above, but ensures that nothing in Dir is interpreted as a command.

  * Command
    One of 
    * homonyms(From)
       attach homonyms From pack
    * init(Lib)
    * init(Lib,Cxt)
       declare initilization call (library can be loaded without this firing, if so needed, as is the
       case for lib_mkindex/1)
    * suggests(Lib)
       it is likely you need Lib for full functionalilty
    * expects(Pid,Mess)
    * expects(Pid,Mess,Call)
       complains if Pid is not defined at loading time. Mess should be a debug style message
       with one ~w which will be called with Pid as its printing argument. 
       If call is present, is called after the printing of the message.
    
Opts
  * index(Idx)
    whether to load indices

  * homonym(Hnym)
    whether to load homonym file-locators

  * load(Load)
    whether to load the main entry point of Repo

  * type(Type)
    enforce a particular type of repository (pack or lib)
  
The defaults depend on whether Repo is a pack or a lib. 
* Packs get defaults
    [load(true),index(true),homonym(false),type(pack)]
* Libs get defaults
    [load(false),index(true),homonym(true),type(lib)]

When invvoked with code attaching operands (SysLibrary, Pack or Lib) 
the predicate will first load anything that needs to be loaded in their native module
and then import predicates from that module. Attaching a lib or pack means that the 
predicates pointed to by indices and by file name from the target pack/lib become
available to the importee. Option =|index(Idx)|= controls whether =LibIndex.pl=
based indices are attahced whereas =|homonym(Hmns)|= control the attachment of
the file names from the target.

For example to only import the interface predicates of pack =ex1= use
==
?- lib(ex1, [type(pack),load(true),index(false),homonym(false)]).
==

Assume that ex1 is a pack that is not installed on your Prolog installation, but you have its sources
unpacked on local dir =|/tmp/ex1/|= you can load it interface predicates with:

For example to only import the interface predicates of pack =ex1= use
==
?- lib('/tmp/ex1', [type(pack),load(true),index(false),homonym(false)]).
==

Assume there is a file =|src/lib/foo.pl|= in =ex1= defining predicate foo/1, then you can load its code with
==
?- lib('/tmp/ex1', [type(pack),load(true),index(false),homonym(true)]).
?- lib(foo/1).
==
The above will first load foo.pl (by means of matching its filename to the predicate name) into ex1:
and then assuming that this loaded ex1:foo/1 it will import it into current context (here this is =user+).

Assuming foo.pl also defines predicate bar/2 and there is a file =|src/LibIndex.pl|= within =ex1= containing the
line
==
lib_index( bar, 2, swipl(_), user, 'lib/foo.pl' ). 
==
Then the code for foo_bar/2 can be loaded with
==
?- lib('/tmp/ex1', [type(pack),load(true),index(true),homonym(false)]).
?- lib(bar/2).
==


Pack lib can be used to create and access *skeleton* packs. These packs, may load very little interface
code but their code base can be loaded on demand and piece-meal. That is if a specific non-interface
predicate is required, it will be located and loaded along with all its dependencies.

An example of such a pack is =|stoics_lib|=. 
The following commands: 1. load the minimal interface, 2, load the code for a specific non-interface predicate.
==
?- lib(stoics_lib).
?- lib(kv_decompose/3).
==

The above two directives can be shortened to:
==
?- lib(stoics_lib:kv_decompose/3).
==

@tbd options that load all indexed

*/
lib( Repo ) :-
    % fixme: add alias() command
    lib( Repo, [] ).

lib( Repo, ArgS ) :-
    lib_loading_context( Cxt ),
    lib_en_list( ArgS, Args ),
    lib( Repo, Cxt, Args ).

lib( Pn/Pa, Cxt, Args ) :-
    !,
    lib_load( Cxt, Pn, Pa, Args ).
lib( Repo:Pn/Pa, Cxt, Args ) :-
    !,
    lib( Repo, Cxt, Args ),
    lib( Pn/Pa, Cxt, [repo(Repo)|Args] ).
lib( homonyms(Repo), _, _Args )  :-   
                  % load local homonyms as coming from Repo. can be added to
                  % pack to indicate that LibIndex is incomplete or missing 
                  % by default packs do not load their homonyms
    !,
    lib_homonyms( Repo ).
lib( alias(Alias), Cxt, Opts ) :-
    !,
    lib_alias( Alias, Cxt, Opts ).
lib( suggests(Lib), _, _Args ) :-  % fixme: add note() option
    !,
    lib_suggests( Lib ).
lib( expects(Lib,Mess), _, _Opts ) :-  % fixme: add note() option 
    !,
    lib_expects( Lib, Mess ).
lib( expects(Lib,Mess,Goal), _, _Opts ) :-  % fixme: add note() option 
    !,
    lib_expects( Lib, Mess, Goal ).
lib( init(Lib), Cxt, _Opts ) :-
    !,
    lib_init( Lib, Cxt ).
lib( Repo, Cxt, Args ) :-
    lib_tables:lib_repo(Repo,Type,Root,Load),
    !,
    lib_repo( Repo, Type, Root, Load, Cxt, Args ).
lib( Root, Cxt, Args ) :-
    lib_tables:lib_repo(Repo,Type,Root,Load),
    !,
    lib_repo( Repo, Type, Root, Load, Cxt, Args ).
lib( Repo, Cxt, Args ) :-
    lib_type( Repo, RepoType, RepoMod, RepoRoot, RepoLoad ),
    !,
    lib_reg_repo( RepoMod, RepoType, RepoRoot, RepoLoad ),
    MsId = 'Identified repo:~w as: ~w, loading in: ~w, with root: ~w',
    debug( lib, MsId, [Repo,RepoType,RepoMod,RepoRoot] ),
    lib_defaults( RepoType, Defs ),
    append( Args, Defs, Opts ),
    lib( RepoMod, RepoRoot, RepoLoad, Cxt, Opts ).
lib( SysLib, Cxt, _Args ) :-
    catch( Cxt:use_module(library(SysLib)), _, fail ),
    !,
    debug( lib, 'System library: ~w, loaded in: ~w', [SysLib,Cxt] ).
lib( Pack, Cxt, Opts ) :-
    lib_missing( Pack, Cxt, Opts ),
    !.
lib( Repo, Cxt, Opts ) :-
    memberchk( mode(Mode), Opts ),
    lib_not_found( Mode, Repo, Cxt ).

lib_not_found( self, Repo, _Cxt ) :-
    Mess = 'Failed to locate repository:~w, (no local lib, local pack or remote pack)',
    lib_message_report( Mess, [Repo], informational ).
lib_not_found( suggests, Repo, _Cxt ) :-
    Mess = 'Failed to locate suggested repository:~w, (no local lib, local pack or remote pack)',
    lib_message_report( Mess, [Repo], informational ).

lib_missing( Pack, Cxt, Opts ) :-
    G = query_pack_server(search(Pack), Result, [] ),
    catch( prolog_pack:G, _Ball, fail ),
    Result \== false,
	!,
    memberchk( mode(Mode), Opts ),
	prolog_pack:confirm( pack_on_server(Mode,Pack), yes, [] ),
	pack_install( Pack ),
	Cxt:use_module( library(Pack) ).

lib_alias( Alias, Cxt, Opts ) :-
    !,
    absolute_file_name( Alias, Dir, [access(exist)] ),
    lib( Dir, Cxt, Opts ).
lib_alias( Alias, _Cxt, _Opts ) :-
    throw( alias_does_not_correspont_to_lib(Alias) ).

lib( Repo, Root, Load, Cxt, Opts ) :-
    Setup = asserta( lib_tables:lib_context(Repo,Root) ),
    Goal  = lib_load_file( Load, Repo, Opts ),
    Clean = ( once(retract(lib_tables:lib_context(Repo,Root))) ),
    setup_call_cleanup(Setup, Goal, Clean),
    findall( _, 
                    (   predicate_property(Repo:Ph,exported),
                        functor(Ph,Pn,Pa),
                        Cxt:import(Repo:Pn/Pa)
                    )
                    ,
                        _ ),
    memberchk( index(IdxB), Opts ),
    lib_attach_indices( IdxB, Root, Repo, Cxt ),
    memberchk( homonym(LocB), Opts ),
    lib_attach_filenames( LocB, Root, Repo, Cxt ).

lib_repo( Repo, Type, Root, Load, Cxt, Args ) :-
    Mess = 'Located in memory repo:~w of type: ~w, loading in: ~w, with root: ~w',
    debug( lib, Mess, [Repo,Type,Repo,Root] ),
    lib_defaults( Type, Defs ),
    append( Args, Defs, Opts ),
    lib( Repo, Root, Load, Cxt, Opts ),
    findall( _, (predicate_property(Repo:Phead,exported),functor(Phead,Pn,Pa),Cxt:import(Repo:Pn/Pa)), _ ).
